/*
 * SWE Ereigniskalender - Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49-341-49067-0
 * Fax: +49-341-49067-15
 * E-Mail: info@bitctrl.de
 */

package de.bsvrz.vew.ereigniskal.data;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.vew.syskal.SystemkalenderGueltigkeit;

/**
 * Definition der verkehrlichen Gültigkeit eines Ereignisses.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class VerkehrlicheGueltigkeit {

    private static class GueltigkeitsBereich {

        public static GueltigkeitsBereich of(Duration anfang, ZeitBezug bezugAnfang, Duration ende,
                ZeitBezug bezugEnde) {

            GueltigkeitsBereich bereich = new GueltigkeitsBereich();

            bereich.dauerAnfang = anfang;
            bereich.bezugAnfang = bezugAnfang;
            bereich.dauerEnde = ende;
            bereich.bezugEnde = bezugEnde;

            return bereich;
        }

        private Duration dauerAnfang = Duration.ZERO;
        private ZeitBezug bezugAnfang = ZeitBezug.vorAnfang;
        private Duration dauerEnde = Duration.ZERO;
        private ZeitBezug bezugEnde = ZeitBezug.nachEnde;

        private List<VerkehrlichesIntervall> bestimmeGueltigkeitsIntervalle(LocalDateTime zeitPunkt,
                SystemkalenderGueltigkeit gueltigkeit,
                ZeitlicheGueltigkeit zeitlicheGueltigkeit) {

            VerkehrlichesIntervall wechselBereich = null;
            List<VerkehrlichesIntervall> bereiche = new ArrayList<>();

            if (gueltigkeit.isZeitlichGueltig()) {
                wechselBereich = ermittleGueltigkeitsBereich(gueltigkeit);
                if (wechselBereich.liegtInnerhalb(gueltigkeit) && !wechselBereich.endsBefore(zeitPunkt)) {
                    bereiche.add(wechselBereich);
                }
            }

            bereiche.addAll(bestimmeRelevanteBereicheFuerVorigeZeitlicheGueltigkeit(zeitPunkt, gueltigkeit,
                    zeitlicheGueltigkeit));
            bereiche.addAll(bestimmeRelevanteBereicheFuerNaechsteZeitlicheGueltigkeit(zeitPunkt, gueltigkeit,
                    zeitlicheGueltigkeit));

            return bereiche;
        }

        private Collection<VerkehrlichesIntervall> bestimmeRelevanteBereicheFuerNaechsteZeitlicheGueltigkeit(
                LocalDateTime zeitPunkt,
                SystemkalenderGueltigkeit gueltigkeit, ZeitlicheGueltigkeit zeitlicheGueltigkeit) {

            Collection<VerkehrlichesIntervall> result = new ArrayList<>();
            VerkehrlichesIntervall wechselBereich;
            SystemkalenderGueltigkeit aktuelleGueltigkeit;

            aktuelleGueltigkeit = zeitlicheGueltigkeit
                    .getNaechstenGueltigkeitsBereich(gueltigkeit.getNaechsterWechsel().getZeitPunkt());
            do {

                if (aktuelleGueltigkeit.isZeitlichGueltig()) {
                    wechselBereich = ermittleGueltigkeitsBereich(aktuelleGueltigkeit);
                    if (wechselBereich.liegtInnerhalb(gueltigkeit)) {
                        result.add(wechselBereich);
                        break;
                    }
                    if (wechselBereich.startsAfter(zeitPunkt)) {
                        aktuelleGueltigkeit = SystemkalenderGueltigkeit.NICHT_GUELTIG;
                        continue;
                    }
                    aktuelleGueltigkeit = zeitlicheGueltigkeit
                            .getNaechstenGueltigkeitsBereich(aktuelleGueltigkeit.getErsterWechsel().getZeitPunkt());
                }
            } while (aktuelleGueltigkeit.isZeitlichGueltig());

            return result;
        }

        private Collection<VerkehrlichesIntervall> bestimmeRelevanteBereicheFuerVorigeZeitlicheGueltigkeit(
                LocalDateTime zeitPunkt,
                SystemkalenderGueltigkeit gueltigkeit, ZeitlicheGueltigkeit zeitlicheGueltigkeit) {

            Collection<VerkehrlichesIntervall> result = new ArrayList<>();
            VerkehrlichesIntervall wechselBereich;
            SystemkalenderGueltigkeit aktuelleGueltigkeit = zeitlicheGueltigkeit
                    .getVorigenGueltigkeitsBereich(gueltigkeit.getErsterWechsel().getZeitPunkt());
            do {
                if (aktuelleGueltigkeit.isZeitlichGueltig()) {
                    wechselBereich = ermittleGueltigkeitsBereich(aktuelleGueltigkeit);
                    if (wechselBereich.endsBefore(zeitPunkt)) {
                        aktuelleGueltigkeit = SystemkalenderGueltigkeit.NICHT_GUELTIG;
                        continue;
                    }
                    if (wechselBereich.enthaelt(zeitPunkt)) {
                        result.add(wechselBereich);
                        break;
                    }
                    aktuelleGueltigkeit = zeitlicheGueltigkeit
                            .getVorigenGueltigkeitsBereich(aktuelleGueltigkeit.getErsterWechsel().getZeitPunkt());
                }
            } while (aktuelleGueltigkeit.isZeitlichGueltig());

            return result;
        }

        private VerkehrlichesIntervall ermittleGueltigkeitsBereich(SystemkalenderGueltigkeit gueltigkeit) {

            if (!gueltigkeit.isZeitlichGueltig()) {
                throw new IllegalStateException(
                        "Zeitliche Gültigkeit ist nicht ermittelbar, "
                                + "da der referenzierte Kalendereintrag ungültig ist");
            }

            LocalDateTime verkehrlicherBeginn = null;
            LocalDateTime verkehrlichesEnde = null;

            switch (bezugAnfang) {
            case nachAnfang:
                verkehrlicherBeginn = gueltigkeit.getErsterWechsel().getZeitPunkt().plus(dauerAnfang);
                break;
            case nachEnde:
                verkehrlicherBeginn = gueltigkeit.getNaechsterWechsel().getZeitPunkt().plus(dauerAnfang);
                break;
            case vorAnfang:
                verkehrlicherBeginn = gueltigkeit.getErsterWechsel().getZeitPunkt().minus(dauerAnfang);
                break;
            case vorEnde:
                verkehrlicherBeginn = gueltigkeit.getNaechsterWechsel().getZeitPunkt().minus(dauerAnfang);
                break;
            default:
                throw new IllegalStateException("Der Bezugstyp: " + bezugAnfang + " wird nicht unterstützt!");
            }

            switch (bezugEnde) {
            case nachAnfang:
                verkehrlichesEnde = gueltigkeit.getErsterWechsel().getZeitPunkt().plus(dauerEnde);
                break;
            case nachEnde:
                verkehrlichesEnde = gueltigkeit.getNaechsterWechsel().getZeitPunkt().plus(dauerEnde);
                break;
            case vorAnfang:
                verkehrlichesEnde = gueltigkeit.getErsterWechsel().getZeitPunkt().minus(dauerEnde);
                break;
            case vorEnde:
                verkehrlichesEnde = gueltigkeit.getNaechsterWechsel().getZeitPunkt().minus(dauerEnde);
                break;
            default:
                throw new IllegalStateException("Der Bezugstyp: " + bezugEnde + " wird nicht unterstützt!");
            }

            return VerkehrlichesIntervall.of(verkehrlicherBeginn, verkehrlichesEnde);
        }

        @Override
        public String toString() {
            return "GueltigkeitsBereich [dauerAnfang=" + dauerAnfang + ", bezugAnfang=" + bezugAnfang + ", dauerEnde="
                    + dauerEnde + ", bezugEnde=" + bezugEnde + "]";
        }
    }

    public static final GueltigkeitsBereich ZERO = GueltigkeitsBereich.of(Duration.ZERO, ZeitBezug.vorAnfang,
            Duration.ZERO, ZeitBezug.nachEnde);

    /**
     * erzeugt eine Instanz der {@link VerkehrlicheGueltigkeit} aus den
     * übergebenen Datenverteiler-Daten.
     * 
     * @param data
     *            die Daten
     * @return die neue Instanz
     */
    public static VerkehrlicheGueltigkeit from(Data data) {
        VerkehrlicheGueltigkeit verkehrlicheGueltigkeit = new VerkehrlicheGueltigkeit();
        Array verkehrArray = data.getArray("VerkehrlicheGültigkeit");
        for (int idx = 0; idx < verkehrArray.getLength(); idx++) {
            Duration dauerAnfang = Duration
                    .ofMillis(verkehrArray.getItem(idx).getTimeValue("ZeitDauerAnfangIntervall").getMillis());
            ZeitBezug bezugAnfang = ZeitBezug
                    .of(verkehrArray.getItem(idx).getUnscaledValue("ZeitBezugAnfangIntervall").intValue());
            Duration dauerEnde = Duration
                    .ofMillis(verkehrArray.getItem(idx).getTimeValue("ZeitDauerEndeIntervall").getMillis());
            ZeitBezug bezugEnde = ZeitBezug
                    .of(verkehrArray.getItem(idx).getUnscaledValue("ZeitBezugEndeIntervall").intValue());
            verkehrlicheGueltigkeit.addBereich(dauerAnfang, bezugAnfang, dauerEnde, bezugEnde);
        }
        return verkehrlicheGueltigkeit;
    }

    private Collection<GueltigkeitsBereich> bereiche = new ArrayList<>();

    /**
     * fügt einen Bereich der verkehrlichen Gültigkeit hinzu.
     * 
     * @param dauerAnfang
     *            die Abweichung des Anfangszeitpunktes der verkehrlichen
     *            Gültigkeit
     * @param bezugAnfang
     *            die Relation des Anfangszeitpunktes der verkehrlichen
     *            Gültigkeit zu einem Bereich der zeitlichen Gültigkeit
     * @param dauerEnde
     *            die Abweichgun des Endzeitpunktes der verkehrlichen Gültigkeit
     * @param bezugEnde
     *            die Relation des Endzeitpunktes der verkehrlichen Gültigkeit
     *            zu einem Bereich der zeitlichen Gültigkeit
     */
    public void addBereich(Duration dauerAnfang, ZeitBezug bezugAnfang, Duration dauerEnde, ZeitBezug bezugEnde) {
        bereiche.add(GueltigkeitsBereich.of(dauerAnfang, bezugAnfang, dauerEnde, bezugEnde));
    }

    /**
     * bestimmt die Intervalle der verkehrlichen Gültigkeit für den Zeitpunkt
     * und die übergebenen Bereich der zeitlichen Gültigkeit.
     * 
     * @param zeitPunkt
     *            der Zeitpunkt
     * @param zeitlicheGueltigkeit
     *            der Bereich der zeitlichen Gültigkeit
     * @return die Liste der Intervalle
     */
    public List<VerkehrlichesIntervall> bestimmeRelevanteBereiche(LocalDateTime zeitPunkt,
            ZeitlicheGueltigkeit zeitlicheGueltigkeit) {

        SystemkalenderGueltigkeit gueltigkeit = zeitlicheGueltigkeit.getGueltigkeit(zeitPunkt);
        List<VerkehrlichesIntervall> intervalle = new ArrayList<>();

        if (bereiche.isEmpty()) {
            if (gueltigkeit.isZeitlichGueltig()) {
                return Collections
                        .singletonList(VerkehrlichesIntervall.of(gueltigkeit.getErsterWechsel().getZeitPunkt(),
                                gueltigkeit.getNaechsterWechsel().getZeitPunkt()));
            }
        }

        for (GueltigkeitsBereich bereich : bereiche) {
            List<VerkehrlichesIntervall> intervall = bereich.bestimmeGueltigkeitsIntervalle(zeitPunkt, gueltigkeit,
                    zeitlicheGueltigkeit);
            intervalle.addAll(intervall);
        }

        return intervalle;
    }

    public boolean isObsolete() {
        for( GueltigkeitsBereich bereich : bereiche) {
            if( !bereich.dauerAnfang.equals(Duration.ZERO)) {
                return false;
            }
            if( !bereich.dauerEnde.equals(Duration.ZERO)) {
                return false;
            }
        }
        return true;
    }

}
