/*
 * SWE Ereigniskalender - Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49-341-49067-0
 * Fax: +49-341-49067-15
 * E-Mail: info@bitctrl.de
 */

package de.bsvrz.vew.ereigniskal.data;

import java.time.LocalDateTime;
import java.util.Objects;

import de.bsvrz.vew.syskal.SystemkalenderGueltigkeit;

/**
 * Definition eines Inervalls der verkehrlichen Gültigkeit.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class VerkehrlichesIntervall {

    private final LocalDateTime start;
    private final LocalDateTime ende;

    static VerkehrlichesIntervall of(LocalDateTime start, LocalDateTime ende) {
        return new VerkehrlichesIntervall(start, ende);
    }

    private VerkehrlichesIntervall(LocalDateTime start, LocalDateTime ende) {
        this.start = start;
        this.ende = ende;
    }

    /**
     * ermittelt, ob der übergebenene Zeitpunkt im Intervall enthalten ist
     * (Anfang inklusiv, Ende exklusiv).
     * 
     * @param zeitPunkt
     *            der Zeitpunkt
     * @return true, wenn der Zeitpunkt enthalten ist
     */
    public boolean enthaelt(LocalDateTime zeitPunkt) {
        return !zeitPunkt.isBefore(start) && zeitPunkt.isBefore(ende);
    }

    /**
     * ermittelt, ob das Intervall innerhalb des übergebenen Gültigkeitsbereichs
     * eines Systemkalendereintrags befindet.
     * 
     * @param gueltigkeit
     *            die Systemkalendergültigkeit
     * @return true, wenn das Intervall innerhalb liegt
     */
    public boolean liegtInnerhalb(SystemkalenderGueltigkeit gueltigkeit) {

        if (!start.isBefore(gueltigkeit.getErsterWechsel().getZeitPunkt())
                && start.isBefore(gueltigkeit.getNaechsterWechsel().getZeitPunkt())) {
            return true;
        }

        if (!ende.isBefore(gueltigkeit.getErsterWechsel().getZeitPunkt())
                && ende.isBefore(gueltigkeit.getNaechsterWechsel().getZeitPunkt())) {
            return true;
        }

        if (start.isBefore(gueltigkeit.getErsterWechsel().getZeitPunkt())
                && !ende.isBefore(gueltigkeit.getNaechsterWechsel().getZeitPunkt())) {
            return true;
        }

        return false;
    }

    @Override
    public String toString() {
        return "VerkehrlichesIntervall [start=" + start + ", ende=" + ende + "]";
    }

    /**
     * ermittelt, ob das Intervall zu dem oder nach dem übergebenen Zeitpunkt
     * beginnt.
     * 
     * @param zeitPunkt
     *            der zeitPunkt
     * @return das Ergebnis der Prüfung
     */
    public boolean folgtAuf(LocalDateTime zeitPunkt) {
        return !zeitPunkt.isAfter(start);
    }

    LocalDateTime getStart() {
        return start;
    }

    LocalDateTime getEnde() {
        return ende;
    }

    /**
     * ermittelt, ob das Intervall nach dem übergebenen Zeitpunkt beginnt.
     * 
     * @param zeitPunkt
     *            der zeitPunkt
     * @return das Ergebnis der Prüfung
     */
    public boolean startsAfter(LocalDateTime zeitPunkt) {
        return start.isAfter(zeitPunkt);
    }

    /**
     * ermittelt, ob das Intervall vor dem übergebenen Zeitpunkt beginnt.
     * 
     * @param zeitPunkt
     *            der zeitPunkt
     * @return das Ergebnis der Prüfung
     */
    public boolean startsBefore(LocalDateTime zeitPunkt) {
        return start.isBefore(zeitPunkt);
    }

    /**
     * ermittelt, ob das Intervall nach dem übergebenen Zeitpunkt endet.
     * 
     * @param zeitPunkt
     *            der zeitPunkt
     * @return das Ergebnis der Prüfung
     */
    public boolean endsAfter(LocalDateTime zeitPunkt) {
        return ende.isAfter(zeitPunkt);
    }

    /**
     * ermittelt, ob das Intervall vor dem übergebenen Zeitpunkt endet.
     * 
     * @param zeitPunkt
     *            der zeitPunkt
     * @return das Ergebnis der Prüfung
     */
    public boolean endsBefore(LocalDateTime zeitPunkt) {
        return ende.isBefore(zeitPunkt);
    }

    @Override
    public int hashCode() {
        return Objects.hash(start, ende);
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }

        if (obj == null) {
            return false;
        }

        if (getClass() != obj.getClass()) {
            return false;
        }

        boolean result = Objects.equals(start, ((VerkehrlichesIntervall) obj).start);
        return result && Objects.equals(ende, ((VerkehrlichesIntervall) obj).ende);
    }
}