/*
 * SWE Ereigniskalender - Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49-341-49067-0
 * Fax: +49-341-49067-15
 * E-Mail: info@bitctrl.de
 */

package de.bsvrz.vew.ereigniskal.module;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import de.bsvrz.vew.ereigniskal.data.Ereignis;
import de.bsvrz.vew.ereigniskal.data.KalenderAnfrage;

/**
 * Verwaltung von Ereignissen
 * 
 * Ereignisse werden über ihren Name verwaltet, d.h. die Namen sollten
 * einheitlich sein!
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class EreignisProvider {

    private Map<String, Ereignis> ereignisse = new ConcurrentHashMap<>();

    /**
     * fügt der Verwaltung ein Ereignis hinzu.
     * 
     * @param ereignis
     *            das Ereignis
     */
    public void addEreignis(Ereignis ereignis) {
        ereignisse.put(ereignis.getName(), ereignis);
    }

    /**
     * entfernt das Ereignis mit dem übergebenen Name aus der verwaltung.
     * 
     * @param name
     *            der Name des zu entfernenden Ereignisses
     */
    public void removeEreignis(String name) {
        ereignisse.remove(name);
    }

    /**
     * liefert die Zustandswechsel aller Ereignisse entsprechend der
     * Kalenderanfrage.
     * 
     * Die Kalenderanfrage definiert den Abfragezeitraum und kann Ereignisse
     * filtern.
     * 
     * @param anfrage
     *            die Anfrage
     * @return die Liste der Zustandswechsel
     */
    public List<EreignisZustandsWechsel> getEreignisZustandsWechsel(KalenderAnfrage anfrage) {
        List<EreignisZustandsWechsel> result = new ArrayList<>();

        for (Ereignis ereignis : ereignisse.values()) {
            if (!anfrage.verwendeEreignis(ereignis)) {
                continue;
            }
            result.addAll(
                    ereignis.getEreignisZustandsWechsel(anfrage.getAnfangsZeitpunkt(), anfrage.getEndZeitpunkt()));
        }

        Collections.sort(result, new ZustandsWechselComparator());

        // Filtere Übergänge ohne Änderungen
        Map<Ereignis, EreignisZustandsWechsel> letzteWechsel = new LinkedHashMap<>();
        Iterator<EreignisZustandsWechsel> iterator = result.iterator();
        while (iterator.hasNext()) {
            EreignisZustandsWechsel wechsel = iterator.next();
            EreignisZustandsWechsel letzterWechsel = letzteWechsel.get(wechsel.getEreignis());
            if (letzterWechsel != null && letzterWechsel.getVerkehrlichGueltig() == wechsel.getVerkehrlichGueltig()
                    && letzterWechsel.getZeitlichGueltig() == wechsel.getZeitlichGueltig()) {
                iterator.remove();
            } else {
                letzteWechsel.put(wechsel.getEreignis(), wechsel);
            }
        }

        return result;
    }

    /**
     * liefert das Ereignis mit dem übergebenen Name.
     * 
     * @param name
     *            der Name
     * @return das Ereignis oder null, wenn kein entsprechendes Ereignis
     *         verwaltet wird
     */
    public Ereignis getEreignis(String name) {
        return ereignisse.get(name);
    }
}
