/*
 * SWE Ereigniskalender - Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49-341-49067-0
 * Fax: +49-341-49067-15
 * E-Mail: info@bitctrl.de
 */

package de.bsvrz.vew.ereigniskal.module;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

/**
 * Überwachungstasks, die alle Ereignisse beobachtet und zu den erwarteten
 * Ereigniswechseln eine Neuberechung der Gültigkeiten veranlasst.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class EreignisWechselUeberwacher {

    private LocalDateTime naechsterWechsel;

    private ScheduledExecutorService ueberwachungsExecutor = Executors.newSingleThreadScheduledExecutor();
    private ExecutorService aktualierungsExecutor = Executors.newCachedThreadPool();

    private Map<DavEreignis, EreignisGueltigkeit> publiziert = new ConcurrentHashMap<>();

    private ScheduledFuture<?> aktuellerUeberwacher;

    /**
     * prüft, ob eine Änderung der Gültigkeit bezüglich der letzten
     * Veröffentlichung vorliegt.
     * 
     * @param ereignis
     *            das Ereignis
     * @param gueltigkeit
     *            die neue Gültigkeit
     * @return true, wenn eine Veränderung besteht
     */
    public boolean gueltigKeitGeaendert(DavEreignis ereignis, EreignisGueltigkeit gueltigkeit) {

        EreignisGueltigkeit letzteGueltigkeit = publiziert.get(ereignis);
        if (letzteGueltigkeit == null) {
            return true;
        }

        if (letzteGueltigkeit.isVerkehrlichGueltig() != gueltigkeit.isVerkehrlichGueltig()) {
            return true;
        }

        if (letzteGueltigkeit.isZeitlichGueltig() != gueltigkeit.isZeitlichGueltig()) {
            return true;
        }

        if (!letzteGueltigkeit.getWechselZeit().equals(gueltigkeit.getWechselZeit())) {
            return true;
        }

        return false;
    }

    /**
     * setzt die aktuelle Ereignisgültigkeit und damit die Überwachungzeit für
     * ein Ereignis
     * 
     * @param ereignis
     *            das Ereignis
     * @param gueltigkeit
     *            die Gültigkeit
     */
    public void setGueltigkeit(DavEreignis ereignis, EreignisGueltigkeit gueltigkeit) {
        publiziert.put(ereignis, gueltigkeit);
        setUeberwachungsZeit(gueltigkeit.getWechselZeit());
    }

    private void setUeberwachungsZeit(LocalDateTime wechselZeit) {

        if (naechsterWechsel == null || wechselZeit.isBefore(naechsterWechsel)
                || naechsterWechsel.isBefore(LocalDateTime.now())) {
            if (aktuellerUeberwacher != null) {
                aktuellerUeberwacher.cancel(false);
            }
            naechsterWechsel = wechselZeit;
            long delay = Duration.between(LocalDateTime.now(), wechselZeit).toMillis();
            aktuellerUeberwacher = ueberwachungsExecutor.schedule(() -> aktualisiereEreignisse(), delay,
                    TimeUnit.MILLISECONDS);
        }
    }

    private void aktualisiereEreignisse() {

        Map<DavEreignis, EreignisGueltigkeit> gueltigkeiten = new LinkedHashMap<>(publiziert);
        LocalDateTime now = LocalDateTime.now();
        LocalDateTime neueWechselZeit = now.plusMonths(1);

        for (Entry<DavEreignis, EreignisGueltigkeit> entry : gueltigkeiten.entrySet()) {
            LocalDateTime wechselZeit = entry.getValue().getWechselZeit();
            if (!now.isBefore(wechselZeit)) {
                aktualierungsExecutor.execute(() -> entry.getKey().publiziereStatus());
            } else {
                if (wechselZeit.isBefore(neueWechselZeit)) {
                    neueWechselZeit = wechselZeit;
                }
            }
        }

        setUeberwachungsZeit(neueWechselZeit);
    }
}
