/*
 * SWE Ereigniskalender - Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49-341-49067-0
 * Fax: +49-341-49067-15
 * E-Mail: info@bitctrl.de
 */

package de.bsvrz.vew.ereigniskal.module;

import java.util.List;
import java.util.concurrent.Semaphore;

import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.ereigniskal.data.DavKalenderAnfrage;

/**
 * Instanz einer Task zur Beantwortung einer Kalenderanfrage.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class KalenderAnfrageBeantworter implements Runnable, ClientSenderInterface {

    private static final Debug LOGGER = Debug.getLogger();
    private DavKalenderAnfrage anfrage;
    private EreignisVerwaltung verwaltung;
    private DataDescription antwortDesc;
    private ResultData antwort;
    private Semaphore lock = new Semaphore(0);

    /**
     * erzeugt eine Instanz eines {@link KalenderAnfrageBeantworter} für die
     * übergebene Anfrage und die angegebenen Ereignisverwaltung.
     * 
     * @param anfrage
     *            die Anfrage
     * @param verwaltung
     *            die Verwaltung
     */
    public KalenderAnfrageBeantworter(DavKalenderAnfrage anfrage, EreignisVerwaltung verwaltung) {

        this.anfrage = anfrage;
        this.verwaltung = verwaltung;
        this.antwortDesc = new DataDescription(
                verwaltung.getDav().getDataModel().getAttributeGroup("atg.ereignisKalenderAntwort"),
                verwaltung.getDav().getDataModel().getAspect("asp.antwort"));
    }

    @Override
    public void run() {

        SystemObject anfrager = anfrage.getAnfrager();
        antwort = berechneAntwort();
        try {
            verwaltung.getDav().subscribeSender(this, anfrager, antwortDesc, SenderRole.sender());
        } catch (OneSubscriptionPerSendData e) {
            LOGGER.warning("Unerwartete Fehler bei der Beantwortung einer Kalenderanfrage: " + e.getLocalizedMessage());
        }

        try {
            lock.acquire();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    private ResultData berechneAntwort() {
        
        Data result = verwaltung.getDav().createData(antwortDesc.getAttributeGroup());
        result.getTextValue("absenderZeichen").setText(anfrage.getZeichen());
        result.getUnscaledValue("änderung").setText("Nein");

        LOGGER.fine("EK-Anfrage von " + anfrage.getAnfrager().getNameOrPidOrId() + " für Ereignisse von "
                + anfrage.getAnfangsZeitpunkt() + " bis " + anfrage.getEndZeitpunkt());

        List<EreignisZustandsWechsel> wechselListe = verwaltung.getEreignisZustandsWechsel(anfrage);
        Array array = result.getArray("Ereignis");
        array.setLength(wechselListe.size());

        int idx = 0;
        for (EreignisZustandsWechsel wechsel : wechselListe) {
            Data ereignisItem = array.getItem(idx++);
            ereignisItem.getTimeValue("Zeitpunkt").setMillis(wechsel.getZeitpunktMillis());
            ereignisItem.getReferenceValue("EreignisReferenz").setSystemObject(wechsel.getEreignis().getSystemObject());
            ereignisItem.getUnscaledValue("zeitlichGültig").setText(wechsel.getZeitlichGueltig() ? "Ja" : "Nein");
            ereignisItem.getUnscaledValue("verkehrlichGültig").setText(wechsel.getVerkehrlichGueltig() ? "Ja" : "Nein");
        }

        return new ResultData(anfrage.getAnfrager(), antwortDesc, verwaltung.getDav().getTime(), result);
    }

    @Override
    public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
        if (state == ClientSenderInterface.START_SENDING) {
            try {
                verwaltung.getDav().sendData(antwort);
            } catch (DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
                throw new IllegalStateException("Sollte hier nicht auftreten, da die Sendesteuerung ausgewertet wird.",
                        e);
            }
            lock.release();
        }
    }

    @Override
    public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
        return true;
    }

}
